/*
 * TRESOR Core
 *
 * Copyright (C) 2012 Benjamin Taubmann <taubmann.benjamin@informatik.stud.uni-erlangen.de>
 * Copyright (C) 2013 Johannes Goetzfried <johannes@jgoetzfried.de>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307 USA.
 */

#include "ap.h"
#include "config.h"
#include "initfunc.h"
#include "panic.h"
#include "pcpu.h"
#include "printf.h"
#include "spinlock.h"
#include "sse.h"
#include "string.h"
#include "tresor_asm.h"
#include "tresor.h"

#define MAXCPU      16

#define NOTINIT     0
#define KEYAVAIL    1
#define KEYSET      2

static volatile int             init[MAXCPU];
static volatile unsigned char   key_swap[32];
static volatile unsigned int    init_cpus;
static spinlock_t               lock;

/* If all CPUs are initialized the key must be erased from RAM. */
void clear_swap(void) {
	int cpus = 0;

	spinlock_lock(&lock);
	cpus = ++init_cpus;
	spinlock_unlock(&lock);

	if (cpus > num_of_processors) {
		memset((void *)key_swap, 0, 32);
		asm volatile("wbinvd;");
	}
}


void tresor_init_ap(void) {
	// key has already been set
	if (init[currentcpu->cpunum] == KEYSET) return;

	// waiting for the BSP to be ready
	while (init[currentcpu->cpunum] != KEYAVAIL) ;

	printf("Starting Tresor on CPU %d of %d\n",
	       currentcpu->cpunum, num_of_processors);

	tresor_set_key(key_swap);

	clear_swap();
	init[currentcpu->cpunum] = KEYSET;
}


/* Called directly after entering the password This function parses the
 * configuration and sets the password to the debug register. 
 */
void tresor_init_bsp(void) {
	// Function is called more than once
	// We will set the key only one time!
	static int key_set = NOTINIT;
	int i, j;

	if (key_set != NOTINIT)
		return;

	key_set = KEYAVAIL;
	init_cpus = 0;

	for (i = num_of_processors; i >= 0; i--)
		init[i] = NOTINIT;

	spinlock_init(&lock);

	// Find the tresor key in the configuration
	for (j = 0; j < NUM_OF_STORAGE_KEYS_CONF; j++) {
		if (!strcmp(config.storage.keys_conf[j].crypto_name,"tresor")) {
			int idx = config.storage.keys_conf[j].keyindex;
			u8 *key = config.storage.keys[idx];

			tresor_set_key(key);
			tresor_get_key(key_swap);

			// erase the key from the configuration and clear cache
			memset(key, 0, 32);
			clear_swap();
			asm volatile("wbinvd;");

			printf("Initialized TRESOR on CPU %d of %d\n",
			       currentcpu->cpunum, num_of_processors);

			for (i = num_of_processors; i >= 0; i--)
				init[i] = KEYAVAIL;

			// Take only the first tresor match
			return;
		}
	}

	panic("TRESOR not initialized; key not found\n");
}
