# MARK - User Interface Actions
#
# Copyright (C) 2013  Johannes Goetzfried <johannes@jgoetzfried.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys;
import string;
import subprocess;

import config;
import crypto;
import storage;
import teensy;
import tpm;

from basic import *
from dialog import *


_storage_unsealed  = None;
_storage_cleartext = None;
_teensy            = None;
_tpm               = None;

_password          = u'';

_kek               = b'';
_dek               = b'';

_cdek              = b'';
_iv                = b'';
_salt              = b'';
_hash              = b'';

_token             = b'';
_nonce             =  '';



def mainActionBoot():
	s = StepPanel('Boot existing System', [
		('unseal token and nonce', bootActionUnsealTokenAndNonce),
		('get config',             bootActionConfig),
		('verify nonce',           bootActionVerifyNonce),
		('enter passphrase',       bootActionEnterPassphrase),
		('derive KEK',             bootActionDeriveKEK),
		('decrypt / load DEK',     bootActionLoadDEK),
		('generate new nonce',     bootActionGenerateNonce),
		('seal token and nonce',   bootActionSealTokenAndNonce),
		('boot target system',     bootActionBootTargetSystem)
	], 70);
	s.draw();


def mainActionInstall():
	s = StepPanel('Install new System', [
		('unseal token and nonce', bootActionUnsealTokenAndNonce),
		('get config',             bootActionConfig),
		('verify nonce',           bootActionVerifyNonce),
		('enter passphrase',       bootActionEnterPassphrase),
		('derive KEK',             bootActionDeriveKEK),
		('decrypt / load DEK',     bootActionLoadDEK),
		('generate new nonce',     bootActionGenerateNonce),
		('seal token and nonce',   bootActionSealTokenAndNonce),
		('install new system',     bootActionInstallNewSystem)
	], 70);
	s.draw();


def mainActionSetup():
	s = StepPanel('Configuration or Recovery', [
		('check tpm state',      setupActionCheckTpmState),
		('get config',           setupActionConfig),
		('set new passphrase',   setupActionEnterPassphrase),
		('generate token t',     setupActionGenerateToken),
		('generate nonce n0',    setupActionGenerateNonce),
		('configure device',     setupActionPrepareTeensy),
		('seal token and nonce', setupActionSealTokenAndNonce),
		('generate or set DEK',  setupActionGenerateOrSetDEK),
		('encrypt / store DEK',  setupActionStoreDEK),
		('check configuration',  setupActionCheck)
	], 70);
	s.draw();


def mainActionReboot():
	ui().end();
	subprocess.call('reboot > /dev/null 2>&1', shell=True);
	sys.exit(0);



def setupActionCheckTpmState(panel):
	global _tpm;

	try:
		_tpm = tpm.Tpm();
	except:
		panel.status(COLOR_FAIL, 'tpm in bad state');
		choicebox('Fail', 'TPM is in bad state.\n'
			'Make sure to take ownership of your TPM first!', ['OK']);
		return ACTION_EXIT;

	return ACTION_OK;


def setupActionConfig(panel):
	global _storage_unsealed;
	global _storage_cleartext;

	subprocess.call('mount ' + config.data_dir + ' > /dev/null 2>&1', shell=True);

	_storage_unsealed = storage.Storage(config.config_unsealed);
	_storage_cleartext = storage.Storage(config.config_cleartext);

	try:
		_storage_unsealed.store();
		_storage_cleartext.store();
	except:
		panel.status(COLOR_FAIL, 'getting config failed');
		choicebox('Fail', 'Getting configuration failed.', ['OK']);
		return ACTION_EXIT;

	return ACTION_OK;


def setupActionEnterPassphrase(panel):
	global _password;

	_password = inputbox('Passphrase', 'Enter a new passphrase.', 1, 66, password=True);

	if len(_password) < 8:
		panel.status(COLOR_FAIL, 'passphrase too short');
		return retry('Passphrase must have at least 8 characters.');

	password2 = inputbox('Passphrase', 'Confirm your passphrase.', 1, 66, password=True);

	if password2 != _password:
		panel.status(COLOR_FAIL, 'passphrases do not match');
		return retry('Passphrases do not match.');

	salt = crypto.get_random_bytes(32);
	_storage_cleartext.set('salt', salt);
	_storage_cleartext.set('hash', crypto.hash_passwd(_password.encode('utf-8'), salt));

	return ACTION_OK;


def setupActionGenerateToken(panel):
	global _token;
	_token = crypto.get_random_bytes(32);
	_storage_unsealed.set('token', _token);
	return ACTION_OK;


def setupActionGenerateNonce(panel):
	global _nonce;
	_nonce = crypto.get_random_alphanum(64);
	_storage_unsealed.set('nonce', _nonce);
	return ACTION_OK;


def setupActionPrepareTeensy(panel):
	device = None;

	choicebox('Device Configuration', 'The device will be configured...\n'
		'Press and release the button on your device!\n'
		'Hit OK if you are ready.', ['OK']);

	try:
		teensy.flash();
		teensy.start();
	except:
		panel.status(COLOR_FAIL, 'configuring failed');
		return retry('Configuring the device failed!');

	try:
		device = teensy.Teensy(config.teensy_device);
	except:
		panel.status(COLOR_FAIL, 'communication failed');
		return retry('Communication with the device failed!');

	try:
		device.verify(config.initial_nonce);
	except:
		panel.status(COLOR_FAIL, 'verifying initial nonce failed');
		device.close();
		return retry('Verifying initial nonce failed!');

	try:
		device.store(_nonce);
	except:
		panel.status(COLOR_FAIL, 'storing new nonce failed');
		device.close();
		return retry('Storing new nonce onto the device failed!');

	device.close();

	choicebox('Device Configuration', 'Your device has been configured successfully.\n'
		'The LED on your device should be switched on!\n'
		'Press and release the button on your device again!\n'
		'Hit OK if you are ready.', ['OK']);

	try:
		teensy.start();
	except:
		panel.status(COLOR_FAIL, 'restarting failed');
		return retry('Restarting device failed!');

	return ACTION_OK;


def setupActionSealTokenAndNonce(panel):
	try:
		_storage_unsealed.store();
	except:
		panel.status(COLOR_FAIL, 'writing config failed');
		choicebox('Fail', 'Writing configuration failed.', ['OK']);
		return ACTION_EXIT;

	try:
		_tpm.seal(config.config_sealed, config.config_unsealed);
	except:
		panel.status(COLOR_FAIL, 'sealing failed');
		choicebox('Fail', 'Sealing failed.', ['OK']);
		return ACTION_EXIT;

	return ACTION_OK;


def setupActionGenerateOrSetDEK(panel):
	global _kek;
	global _dek;

	_kek = crypto.derive_key(_password.encode('utf-8'), _token);

	ret = choicebox('Disk Encryption Key', '', [
		'Generate a random DEK',
		'Enter the DEK manually (do this for recovery)',
		'Abort (return to main menu)',
		]);

	if ret == 2:
		return ACTION_EXIT;

	if ret == 0:
		_dek = crypto.get_random_bytes(32);
	else:
		panel.status(COLOR_ACTIVE, 'waiting for DEK');
		dek = inputbox('Disk Encryption Key',
			'Enter the 256 bit DEK in hex (64 characters):\n\n'
			'(Non-hex characters will be ignored.)',
			2, 16 * 3);

		dek = filter(lambda x: x in string.hexdigits, dek);
		if len(dek) != 64:
			panel.status(COLOR_FAIL, 'dek broken');
			return retry('DEK is invalid.');

		_dek = dek.decode('hex');

	dekdump = ' '.join(["%02X" % ord(x) for x in _dek]);
	dekdump = dekdump[:47] + '\n' + dekdump[48:];

	ret = choicebox('Disk Encryption Key', 'DEK OK?\n\n' + dekdump +
		'\n\nATTENTION: You need this key for recovery.\n'
		'Write it down and store it at a secure location.\n'
		'(This is your only chance to do so!)', [
		'Yes',
		'No, enter/generate a new one',
		'No, abort (return to main menu)'
		]);

	if ret == 2:
		return ACTION_EXIT;

	if ret == 1:
		return ACTION_REPEAT;

	return ACTION_OK;


def setupActionStoreDEK(panel):
	iv = crypto.get_random_bytes(16);
	cipher = crypto.Cipher(_kek, iv);

	_storage_cleartext.set('iv', iv);
	_storage_cleartext.set('cdek', cipher.encrypt(_dek));

	try:
		_storage_cleartext.store();
	except:
		panel.status(COLOR_FAIL, 'writing dek failed');
		choicebox('Fail', 'Writing DEK failed.', ['OK']);
		return ACTION_EXIT;

	return ACTION_OK;


def setupActionCheck(panel):
	subprocess.call('umount ' + config.data_dir + ' > /dev/null 2>&1', shell=True);

	panel.status(COLOR_DONE, 'done');
	choicebox('Success', 'All steps were completed successfully.', ['OK']);
	return ACTION_OK;



def bootActionUnsealTokenAndNonce(panel):
	global _tpm;

	subprocess.call('mount ' + config.data_dir + ' > /dev/null 2>&1', shell=True);

	try:
		_tpm = tpm.Tpm();
	except:
		panel.status(COLOR_FAIL, 'tpm in bad state');
		choicebox('Fail', 'TPM is in bad state.\n'
			'Make sure to take ownership of your TPM first', ['OK']);
		return ACTION_EXIT;

	try:
		_tpm.unseal(config.config_unsealed, config.config_sealed);
	except:
		panel.status(COLOR_FAIL, 'unsealing failed');
		choicebox('Fail', 'Unsealing failed.', ['OK']);
		return ACTION_EXIT;

	return ACTION_OK;


def bootActionConfig(panel):
	global _storage_cleartext;
	global _storage_unsealed;

	global _cdek;
	global _iv;
	global _salt;
	global _hash;
	global _token;
	global _nonce;

	_storage_cleartext = storage.Storage(config.config_cleartext);
	_storage_unsealed = storage.Storage(config.config_unsealed);

	try:
		_storage_cleartext.load();
		_storage_unsealed.load();

		_cdek  = _storage_cleartext.get('cdek');
		_iv    = _storage_cleartext.get('iv');
		_salt  = _storage_cleartext.get('salt');
		_hash  = _storage_cleartext.get('hash');

		_token = _storage_unsealed.get('token');
		_nonce = _storage_unsealed.get('nonce');
	except:
		panel.status(COLOR_FAIL, 'getting config failed');
		choicebox('Fail', 'Getting configuration failed.', ['OK']);
		return ACTION_EXIT;

	return ACTION_OK;


def bootActionVerifyNonce(panel):
	global _teensy;

	try:
		_teensy = teensy.Teensy(config.teensy_device);
	except:
		panel.status(COLOR_FAIL, 'communication failed');
		return retry('Communication with the device failed.');

	try:
		_teensy.verify(_nonce);
	except:
		panel.status(COLOR_FAIL, 'verifying nonce failed');
		ret = choicebox('Fail', 'Verifying nonce failed.\n'
			'Your system might be compromised!\n'
			'Only boot your system if you know what you are doing.',
			['Abort (recommended)', 'Proceed anyway (NOT recommended)']);
		if ret == 0:
			return ACTION_EXIT;

	return ACTION_OK;


def bootActionEnterPassphrase(panel):
	global _password;

	_password = inputbox('Passphrase', 'Enter your passphrase.', 1, 66, password=True);

	if _hash != crypto.hash_passwd(_password.encode('utf-8'), _salt):
		panel.status(COLOR_FAIL, 'invalid password');
		return retry('Invalid password!');

	return ACTION_OK;


def bootActionDeriveKEK(panel):
	global _kek;
	_kek = crypto.derive_key(_password.encode('utf-8'), _token);
	return ACTION_OK;


def bootActionLoadDEK(panel):
	global _dek;
	cipher = crypto.Cipher(_kek, _iv);
	_dek = cipher.decrypt(_cdek);
	return ACTION_OK;


def bootActionGenerateNonce(panel):
	global _nonce;

	nonce = crypto.get_random_alphanum(64);

	try:
		_teensy.store(nonce);
	except:
		panel.status(COLOR_FAIL, 'storing new nonce failed');
		_teensy.close();
		ret = choicebox('Fail', 'Storing new nonce to the device failed.\n'
			'You should abort and check your device!',
			['Abort (recommended)', 'Proceed anyway (NOT recommended)']);
		if ret == 0:
			return ACTION_EXIT;
		else:
			return ACTION_OK;

	_nonce = nonce;
	_teensy.close();

	return ACTION_OK;


def bootActionSealTokenAndNonce(panel):
	_storage_unsealed.set('nonce', _nonce);

	try:
		_storage_unsealed.store();
	except:
		panel.status(COLOR_FAIL, 'writing config failed');
		choicebox('Fail', 'Writing configuration failed.', ['OK']);
		return ACTION_EXIT;

	try:
		_tpm.seal(config.config_sealed, config.config_unsealed);
	except:
		panel.status(COLOR_FAIL, 'sealing failed');
		choicebox('Fail', 'Sealing failed.', ['OK']);
		return ACTION_EXIT;

	return ACTION_OK;


def bootActionBootTargetSystem(panel):
	subprocess.call('umount ' + config.data_dir + ' > /dev/null 2>&1', shell=True);

	panel.status(COLOR_DONE, 'done');
	ret = choicebox('Success', 'All steps were completed successfully.\n'
		'Please remove the USB device now!', ['OK']);

	ui().end();

	p = subprocess.Popen(['/opt/minios_boot'], stdin=subprocess.PIPE);
	p.communicate(input=_dek);

	sys.exit(0);
	return ACTION_REPEAT;


def bootActionInstallNewSystem(panel):
	subprocess.call('umount ' + config.data_dir + ' > /dev/null 2>&1', shell=True);

	panel.status(COLOR_DONE, 'done');
	ret = choicebox('Success', 'All steps were completed sucessfully.\n'
		'Please remove the USB device now!\n', ['OK']);

	panel.status(COLOR_ACTIVE, "processing");
	choicebox('Installation', 'Insert your installation media now!\n'
		'Only bootable USB sticks are supported!\n'
		'Press OK if you are ready.', ['OK']);

	p = subprocess.Popen('ls /dev/sd?', shell=True, stdout=subprocess.PIPE);
	out = p.communicate()[0][:-1];
	dev = out.split('\n');
	devinfo = ['none'];

	for i in dev:
		p = subprocess.Popen('fdisk -l ' + i + ' 2> /dev/null | grep \'^Disk /\'',
				     shell=True, stdout=subprocess.PIPE);
		out = p.communicate()[0];
		devinfo.append(out.split('\n')[0]);

	src = choicebox('Installation', 'Choose the installation source device.', devinfo);
	dst = choicebox('Installation', 'Choose the installation target device.', devinfo);

	if src == 0 or dst == 0 or src == dst:
		choicebox('Installation', 'Invalid source or target device!',
			['OK']);
		return ACTION_EXIT;

	ret = choicebox('Installation', 'All current data on the device\n'
		'' + devinfo[dst] + '\n'
		'will be removed!\n'
		'Copying may take some time...\n'
		'Do you want to proceed?',
		['Proceed', 'Abort (return to main menu)']);

	if ret == 1:
		return ACTION_EXIT;

	# copy install image: max. 4GB are copied
	panel.status(COLOR_ACTIVE, "copying data...");
	subprocess.call('dd if=\"' + dev[src-1] + '\" of=\"' + dev[dst-1] + '\" '
			'bs=4M count=1024 > /dev/null 2>&1', shell=True);

	# wipe disklabel (some installers complain if the target device has a disklabel)
	subprocess.call('wipefs -a ' + dev[dst-1] + ' > /dev/null 2>&1', shell=True);

	# encrypt install image: max. 4GB are encrypted
	panel.status(COLOR_ACTIVE, "encrypting data...");
	p = subprocess.Popen(['/opt/minios_crypt', dev[dst-1], dev[dst-1], '8388608'],
			     stdin=subprocess.PIPE);
	p.communicate(input=_dek);

	panel.status(COLOR_DONE, 'done');
	choicebox('Installation', 'Copying has been finished.\n'
		'Please do NOT unplug your installation media.\n'
		'Your system will now boot into the installation!', ['OK']);

	ui().end();

	p = subprocess.Popen(['/opt/minios_boot'], stdin=subprocess.PIPE);
	p.communicate(input=_dek);

	sys.exit(0);
	return ACTION_REPEAT;

